import logging, json, sys, hashlib

if (sys.version_info > (3, 0)):
    def long(value):
        return int(value)

from functools import wraps

from redis import StrictRedis
from pymemcache.client.base import Client as MemcachedClient

__version__ = '0.2.1'

class CachualCache(object):
    """Base class for all cache implementations. Provides the
    :meth:`~CachualCache.cached` decorator which can be applied to methods
    whose return value you want to cache. This class should not be used
    directly, and is meant to be subclassed.

    Subclasses should define a **get** method, which takes a single string
    argument, and returns the value in the cache for that key, or ``None`` in
    the case of a cache miss; and a **put** method, which takes three arguments
    for the cache key, the value to store, and a TLL (which may be none) and
    puts the value in the cache.
    """
    def __init__(self):
        self.logger = logging.getLogger("cachual")

    def cached(self, ttl=None, pack=None, unpack=None):
        """Functions decorated with this will have their return values cached.
        It should be used as follows::

            cache = RedisCache()

            @cache.cached()
            def method_to_be_cached(arg1, arg2, arg3='default'):
                ...

        A unique cache key will be generated for each function call, so that
        different values for the arguments will result in different cache keys.
        When you decorate a function with this, the cache will be checked
        first; if there is a cache hit, the value in the cache will be
        returned. Otherwise, the function is executed and the return value is
        put into the cache.

        Cache get/put failures are logged but ignored; if the cache goes down,
        the function will continue to execute as normal.

        :type ttl: integer
        :param ttl: The time-to-live in seconds. For caches that support TTLs,
                    the keys will expire after this time. If None (the
                    default), the cache default will be used (usually no
                    expiration).

        :type pack: function
        :param pack: If specified, this function will be called with the
                     decorated function's return value, and the result will be
                     stored in the cache. This can be used to alter the value
                     that actually gets stored in the cache in case you need to
                     process it first (e.g. dump a JSON string that is properly
                     escaped).

        :type unpack: function
        :param unpack: If specified, this function will be called with the
                       value from the cache in the event of a cache hit, and
                       the result will be returned to the caller. This can be
                       used to alter the value that gets returned from a cache
                       hit, in case you need to process it first (e.g. turn a
                       JSON string into a Python dictionary).
        """
        def decorator(f):
            @wraps(f)
            def decorated(*args, **kwargs):
                key = self._get_key_from_func(f, args, kwargs)
                self.logger.debug("key: [%s]" % key)
                try:
                    value = self.get(key)
                    if value is not None:
                        self.logger.debug("got value from cache: %s" % value)
                        if unpack is not None:
                            return unpack(value)
                        return value
                except:
                    self.logger.warn("Error getting value", exc_info=1)

                self.logger.debug("no value from cache, calling function")
                value = f(*args, **kwargs)
                self.logger.debug("got value from function call: %s" % value)
                try:
                    packed = value if pack is None else pack(value)
                    self.put(key, packed, ttl)
                except:
                    self.logger.warn("Error putting value", exc_info=1)
                return value
            return decorated
        return decorator

    def _get_key_from_func(self, f, args, kwargs):
        """Internal function to build the cache key from the function and the
        args and kwargs it was called with."""
        if args:
            args_str = ', '.join([_unicode(a) for a in args])
        else:
            args_str = None
        if kwargs:
            kwargs_str = ', '.join(\
                    ['%s=%s' % (_unicode(k), _unicode(kwargs[k]))\
                    for k in sorted(kwargs.keys())])
        else:
            kwargs_str = None

        prefix = f.__module__ + '.' + f.__name__
        suffix = ''
        if args_str is not None:
            suffix = '(' + args_str
            if kwargs_str is not None:
                suffix = suffix + ', ' + kwargs_str + ')'
            else:
                suffix = suffix + ')'
        else:
            if kwargs_str is not None:
                suffix = '(' + kwargs_str + ')'
            else:
                suffix = '()'

        m = hashlib.md5()
        key = ('%s%s' % (prefix, suffix)).encode('utf-8')
        m.update(key)
        return m.hexdigest()

class RedisCache(CachualCache):
    """A cache using `Redis <https://redis.io/>`_ as the backing cache. All
    values will be stored as strings, meaning if you try to store non-string
    values in the cache, their unicode equivalent will be stored. If you want
    to alter this behavior e.g. to store Python dictionaries, use the
    pack/unpack arguments when you specify your @cached decorator.

    :type host: string
    :param host: The Redis host to use for the cache.

    :type port: integer
    :param port: The port to use for the Redis server.

    :type db: integer
    :param db: The Redis database to use on the server for the cache.

    :type kwargs: dict
    :param kwargs: Any additional args to pass to the :class:`CachualCache`
                   constructor.
    """
    def __init__(self, host='localhost', port=6379, db=0, **kwargs):
        super(RedisCache, self).__init__(**kwargs)
        self.client = StrictRedis(host=host, port=port, db=db)

    def get(self, key):
        """Get a value from the cache using the given key.

        :type key: string
        :param key: The cache key to get the value for.

        :returns: The value for the cache key, or None in the case of cache
                  miss.
        """
        return self.client.get(key)

    def put(self, key, value, ttl=None):
        """Put a value into the cache at the given key. If the value is not a
        string, its unicode value will be used. This behavior is defined by the
        underlying Redis library, and could be subject to change in future
        versions; thus it is safest to only store strings in Redis (although
        basic types should serialize into a string cleanly, you will need to
        unpack them when they are retrieved from the cache).

        :type key: string
        :param key: The cache key to use for the value.

        :param value: The value to store in the cache.

        :type ttl: integer
        :param ttl: The time-to-live for key in seconds, after which it will
                    expire.
        """
        self.client.set(key, value, ex=ttl)

class MemcachedCache(CachualCache):
    """A cache using `Memcached <https://memcached.org/>`_ as the backing
    cache. The same caveats apply to keys and values as for Redis - you should
    only try to store strings (using the packing/unpacking functions). See the
    documentation on Keys and Values here:
    :class:`pymemcache.client.base.Client`.

    :type host: string
    :param host: The Memcached host to use for the cache.

    :type port: integer
    :param port: The port to use for the Memcached server.

    :type kwargs: dict
    :param kwargs: Any additional args to pass to the :class:`CachualCache`
                   constructor.
    """
    def __init__(self, host='localhost', port=11211, **kwargs):
        super(MemcachedCache, self).__init__(**kwargs)
        self.client = MemcachedClient((host, port))

    def get(self, key):
        """Get a value from the cache using the given key.

        :type key: string
        :param key: The cache key to get the value for.

        :returns: The value for the cache key, or None in the case of cache
                  miss.
        """
        return self.client.get(key)

    def put(self, key, value, ttl=None):
        """Put a value into the cache at the given key. For constraints on keys
        and values, see :class:`pymemcache.client.base.Client`.

        :type key: string
        :param key: The cache key to use for the value.

        :param value: The value to store in the cache.

        :type ttl: integer
        :param ttl: The time-to-live for key in seconds, after which it will
                    expire.
        """
        if ttl is None:
            ttl = 0
        self.client.set(key, value, expire=ttl)

def pack_json(value):
    """Pack the given JSON structure for storage in the cache by dumping it as
    a JSON string.

    :param value: The JSON structure (e.g. dict, list) to pack.

    :rtype: string
    :returns: The JSON structure as a JSON string.
    """
    return json.dumps(value)

def unpack_json(value):
    """Unpack the given string by loading it as JSON.

    :type value: string
    :param value: The string to unpack.

    :returns: The string as JSON.
    """
    return json.loads(value)

def unpack_json_python3(value):
    """Unpack the given Python 3 bytes by loading them as JSON.

    :type value: bytes
    :param value: The bytes to unpack.

    :returns: The bytes as JSON.
    """
    return unpack_json(value.decode('utf-8'))

def unpack_bytes(value):
    """Unpack the given Python 3 bytes by loading them as a Python 3 unicode
    string, assuming UTF-8 encoding.

    :type value: bytes
    :param value: The bytes to unpack.

    :returns: The bytes as a Python 3 unicode string, assuming UTF-8 encoding.
    """
    return value.decode('utf-8')

def unpack_int(value):
    """Unpack the given string into an integer.

    :type value: string
    :param value: The string to unpack.

    :rtype: integer
    :returns: The string as an integer.
    """
    return int(value)

def unpack_long(value):
    """Unpack the given string into a long.

    :type value: string
    :param value: The string to unpack.

    :rtype: long
    :returns: The string as a long.
    """
    return long(value)

def unpack_float(value):
    """Unpack the given string into a float.

    :type value: string
    :param value: The string to unpack.

    :rtype: float
    :returns: The string as a float.
    """
    return float(value)

def unpack_bool(value):
    """Unpack the given string into a boolean. 'True' will become True, and
    'False' will become False (as per the unicode values of booleans); anything
    else will result in a ValueError.

    :type value: string
    :param value: The string to unpack.

    :rtype: bool
    :returns: The string as a boolean.
    """
    if value == 'True':
         return True
    elif value == 'False':
         return False
    raise ValueError("Cannot convert %s to bool" % value)

def _unicode(value):
    """Helper function to return the value as a unicode, regardless of the
    Python version of the type of the value."""
    if (sys.version_info > (3, 0)): # Handling for Python 3; str is unicode
        return str(value)

    if isinstance(value, unicode):
        return value
    return unicode(str(value), encoding='utf-8')



from cachual import (pack_json, unpack_json, unpack_int, unpack_long,
                     unpack_float, unpack_bool, unpack_json_python3,
                     unpack_bytes)

import sys
if (sys.version_info > (3, 0)):
    def unicode(value):
        return str(value)

    def long(value):
        return int(value)

from mock import MagicMock, mock

import pytest, sys

@mock.patch("cachual.json")
def test_pack_json(mock_json):
    test_value = "testing"
    mock_json.dumps = MagicMock(return_value=test_value)
    
    assert pack_json("test") == test_value
    mock_json.dumps.assert_called_with("test")

@mock.patch("cachual.json")
def test_unpack_json(mock_json):
    test_value = "testing"
    mock_json.loads = MagicMock(return_value=test_value)

    assert unpack_json("test") == test_value
    mock_json.loads.assert_called_with("test")

def test_unpack_json_python3():
    if (sys.version_info > (3, 0)):
        test_value = bytes('{"test": "testing"}', encoding="utf-8")
    else:
        test_value = str('{"test": "testing"}')
    assert unpack_json_python3(test_value) == {"test": "testing"}

def test_unpack_bytes():
    if (sys.version_info > (3, 0)):
        test_value = bytes("testing", encoding="utf-8")
    else:
        test_value = str("testing")
    assert unpack_bytes(test_value) == "testing"

def test_unpack_int():
    test_value = "3"
    assert unpack_int(test_value) == int(test_value)

def test_unpack_long():
    test_value = "3"
    assert unpack_long(test_value) == long(test_value)

def test_unpack_float():
    test_value = "3.2"
    assert unpack_float(test_value) == float(test_value)

def test_unpack_bool_true():
    test_value = 'True'
    assert unpack_bool(test_value) == True

def test_unpack_bool_false():
    test_value = 'False'
    assert unpack_bool(test_value) == False

def test_unpack_bool_invalid():
    test_value = 'Invalid'
    with pytest.raises(ValueError):
        unpack_bool(test_value)



from cachual import MemcachedCache

from mock import MagicMock, mock

@mock.patch('cachual.MemcachedClient')
def test_ctor(mock_memcached):
    host = "host"
    port = 1234

    mock_memcached.return_value = "test"
    unit = MemcachedCache(host, port)

    mock_memcached.assert_called_with((host, port))
    assert unit.client == "test"

@mock.patch('cachual.MemcachedClient')
def test_get(mock_memcached):
    client = MagicMock()
    client.get = MagicMock()
    mock_memcached.return_value = client
    key = "test"

    unit = MemcachedCache()
    unit.get(key)
    client.get.assert_called_with(key)

@mock.patch('cachual.MemcachedClient')
def test_put(mock_memcached):
    client = MagicMock()
    client.set = MagicMock()
    mock_memcached.return_value = client
    key = "test"
    value = "value"
    ttl = 5

    unit = MemcachedCache()
    unit.put(key, value, ttl)
    client.set.assert_called_with(key, value, expire=ttl)

@mock.patch('cachual.MemcachedClient')
def test_put_no_ttl(mock_memcached):
    client = MagicMock()
    client.set = MagicMock()
    mock_memcached.return_value = client
    key = "test"
    value = "value"

    unit = MemcachedCache()
    unit.put(key, value)
    client.set.assert_called_with(key, value, expire=0)



from cachual import CachualCache

from mock import MagicMock

import logging, sys

if (sys.version_info > (3, 0)):
    def unicode(value):
        return str(value)

    def long(value):
        return int(value)

KEY = "testKey"

def get_unit():
    unit = CachualCache()
    unit.put = MagicMock()
    unit.get = MagicMock()
    unit._get_key_from_func = MagicMock(return_value=KEY)

    logger = logging.getLogger("cachual")
    logger.addHandler(logging.StreamHandler(sys.stderr))
    logger.setLevel("DEBUG")

    return unit

def test_cache_get():
    unit = get_unit()
    test_value = "testing"
    unit.get.return_value = test_value

    @unit.cached()
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)

def test_cache_miss():
    unit = get_unit()
    unit.get.return_value = None
    test_value = "testing"
    
    @unit.cached()
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)
    unit.put.assert_called_with(KEY, test_value, None)

def test_cache_get_error():
    unit = get_unit()
    unit.get.side_effect = Exception("test")
    test_value = "testing"

    @unit.cached()
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)
    unit.put.assert_called_with(KEY, test_value, None)

def test_cache_put_error():
    unit = get_unit()
    test_value = "testing"
    unit.get.return_value = None
    unit.put.side_effect = Exception("test")

    @unit.cached()
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)
    unit.put.assert_called_with(KEY, test_value, None)

def test_cache_put():
    unit = get_unit()
    test_value = "testing"
    unit.get.return_value = None

    @unit.cached()
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)
    unit.put.assert_called_with(KEY, test_value, None)

def test_cache_put_ttl():
    unit = get_unit()
    test_value = "testing"
    ttl = 15
    unit.get.return_value = None

    @unit.cached(ttl=ttl)
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)
    unit.put.assert_called_with(KEY, test_value, ttl)

def test_cache_get_unpack():
    unit = get_unit()
    test_cache_get = "testingCacheGet"
    test_value = "testing"

    unit.get.return_value = test_cache_get
    unpack = MagicMock()
    unpack.return_value = "testing"

    @unit.cached(unpack=unpack)
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)
    unpack.assert_called_with(test_cache_get)

def test_cache_get_unpack_error():
    unit = get_unit()
    test_cache_get = "testingCacheGet"
    test_value = "testing"

    unit.get.return_value = test_cache_get
    unpack = MagicMock()
    unpack.side_effect = Exception("test")

    @unit.cached(unpack=unpack)
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)
    unpack.assert_called_with(test_cache_get)
    unit.put.assert_called_with(KEY, test_value, None)

def test_cache_put_pack():
    unit = get_unit()
    test_value = "testing"
    pack_return_value = "packed"

    unit.get.return_value = None 
    pack = MagicMock()
    pack.return_value = pack_return_value

    @unit.cached(pack=pack)
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)
    pack.assert_called_with(test_value)
    unit.put.assert_called_with(KEY, pack_return_value, None)

def test_cache_put_pack_error():
    unit = get_unit()
    test_value = "testing"
    pack_return_value = "packed"

    unit.get.return_value = None 
    pack = MagicMock()
    pack.side_effect = Exception("test")

    @unit.cached(pack=pack)
    def test(a):
        return a

    assert test(test_value) == test_value
    unit.get.assert_called_with(KEY)
    pack.assert_called_with(test_value)
    unit.put.assert_has_calls([])



# -*- coding: utf-8 -*-
from cachual import CachualCache

from mock import MagicMock

import hashlib, sys

def get_unit():
    return CachualCache()

def get_hash(value):
    m = hashlib.md5()
    m.update(value.encode('utf-8'))
    return m.hexdigest()

def test_no_args():
    def test():
        pass
    test.__module__ = 'test.module'

    key = get_unit()._get_key_from_func(test, None, None)
    assert key == get_hash('test.module.test()')

def test_positional_args():
    def test(a, b):
        pass
    test.__module__ = 'test.module'

    key = get_unit()._get_key_from_func(test, [1, 2], None)
    assert key == get_hash('test.module.test(1, 2)')

def test_kwargs():
    def test(a, b):
        pass
    test.__module__ = 'test.module'

    key = get_unit()._get_key_from_func(test, None,
            {'a': "test1", 'b': "test2"})
    assert key == get_hash('test.module.test(a=test1, b=test2)')

def test_positional_and_kwargs():
    def test(a, b):
        pass
    test.__module__ = 'test.module'

    key = get_unit()._get_key_from_func(test, ['test1'], {'b': "test2"})
    assert key == get_hash('test.module.test(test1, b=test2)')

def test_unicode():
    def test(a):
        pass
    test.__module__ = 'test.module'

    utf8_str = '«ταБЬℓσ»'
    if (sys.version_info > (3, 0)): # Handling for Python 3; str is unicode
        uni = utf8_str
    else:
        uni = utf8_str.decode('utf-8')

    key = get_unit()._get_key_from_func(test, [utf8_str], None)
    assert key == get_hash('test.module.test(%s)' % uni)



"""
Cachual
-------

Cachual is a simple library that makes it easy to cache the return values from
your python functions with an annotation::

    cache = RedisCache()
    
    @cache.cached(ttl=900)
    def is_user_valid(user_id):
        ...
"""
import ast, re

from setuptools import setup

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('cachual.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

setup(
    name='Cachual',
    version=version,
    url='https://github.com/bal2ag/cachual',
    license='BSD',
    author='Alex Landau',
    author_email='balexlandau@gmail.com',
    description='Simple annotation-based caching for your functions',
    long_description=__doc__,
    py_modules=['cachual'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'pymemcache>=1.4.0',
        'redis>=2.10',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)
